

#include "Manipulator.h"

#include <math.h>


#define BFAC 0.11
#define GFAC 0.59
#define RFAC 0.30

#define LINEAR_SCALE


double rand1() {
	double rout;
	rout = rand() / RAND_MAX;
	return(rout);
}


void Manipulator::Setup() {
	ReadInIter = AllSamples.end();
	WriteOutIter = AllSamples.end();
	dispWrite = FALSE;
	DispIter = &ReadInIter;
	ForceBrightness = FALSE;
	MarkMinPix = FALSE;
	last_framenum = 0;
	WriteFilmedTime = (double)0.0;
	WriteVideoTime = (double)0.0;

	// Set up with standard values
	FilmRecordFPS = (double)24.0;
	FilmPlaybackFPS = (double)6.0;
	VideoRecordFPS = (double)29.97;
	VideoPlaybackFPS = (double)29.97;
	MinPixBrightness = (double)100.0;
	GradientReach = 12;
	BrightnessRank = 2;
	DesiredBrightness = 165.0;
	AdjustParameters();
}

Manipulator::Manipulator() {
	mWriter = NULL;
	srand((unsigned int)this);
	Setup();
}

Manipulator::Manipulator(MovieWriter *writer_in) {
	mWriter = writer_in;
	srand((unsigned int)writer_in^0xFFFFFFFF);
	Setup();
}

Manipulator::~Manipulator() {
}


void Manipulator::AdjustParameters() {
	FilmTimePerRecordedFrame = FilmPlaybackFPS / (FilmRecordFPS*VideoRecordFPS);
	VideoTimePerWrittenFrame = (double)1.0 / VideoRecordFPS;
	
	FramesNeededForConsideration = (int)((VideoRecordFPS / FilmPlaybackFPS) + 0.5);
	FramesToSkipBeforeConsideration = (FramesNeededForConsideration + 1) / 2;
}

double Manipulator::CalcPixBrightness(BYTE *bmap) {
	double	pix=0.0;
	double	r,g,b;

	b = (double)*bmap++;
	g = (double)*bmap++;
	r = (double)*bmap++;
	pix = RFAC*r + GFAC*g + BFAC*b;
	return(pix);
}

double Manipulator::CalcBrightness(BITMAPINFO *bmapinfo, BYTE *bmap, double *darkp) {
	int		numpix,i;
	double	b=(double)0.0,p,lastp;
	int		numused=0;

	numpix = bmapinfo->bmiHeader.biHeight*bmapinfo->bmiHeader.biWidth;
	lastp=MinPixBrightness+1;

	for (i=0; i<numpix; i++) {
		p = CalcPixBrightness(bmap);
		if (p>MinPixBrightness) {
			b += p;
			numused++;
			if (MarkMinPix && (lastp<=MinPixBrightness)) {
				*(bmap-3) = 255;
				*(bmap-2) = 0;
				*(bmap-1) = 0;
			}
		}
		else if (MarkMinPix && (lastp>MinPixBrightness)) {
			*bmap = 0;
			*(bmap+1) = 0;
			*(bmap+2) = 255;
		}
		bmap += 3;
		lastp = p;
	}
	if (numused > 0) b /= (double)numused;
	if (darkp != NULL) *darkp = ((double)(numpix-numused)/(double)numpix);
	return(b);
}

void Manipulator::WriteFrame(SIter towrite, double timewrite) {
	mWriter->SaveFrame(towrite->sample, timewrite);
}

void Manipulator::ModifyBitmap(SIter modframe) {
	if (ForceBrightness) {
		BYTE	*ptr=modframe->map;
		double	adjusted_brightness;
		double	scaling,oldb,oldpb;
		double	check1,check2,delta;
		int		i,j,newval,numpix,numused;
		double	scaleadd;
		int		scaleaddR,scaleaddG,scaleaddB;

		oldb = modframe->brightness;
		modframe->brightness_after = 0.0;
		adjusted_brightness = modframe->brightness - (modframe->percent_dark * DarknessFactor);
		scaling = (DesiredBrightness / adjusted_brightness);
		scaleadd = DesiredBrightness - adjusted_brightness;
		scaleaddR = (int)(scaleadd*RFAC);
		scaleaddG = (int)(scaleadd*GFAC);
		scaleaddB = (int)(scaleadd*BFAC);
		numpix = modframe->info->bmiHeader.biHeight*modframe->info->bmiHeader.biWidth;
		numused = 0;
		check1 = modframe->brightness;
		check2 = CalcBrightness(modframe->info, modframe->map, &delta);
		check2 = modframe->percent_dark;
		check1 = 0.0;
		for (i=0; i<numpix; i++) {
			oldpb = CalcPixBrightness(ptr);
#ifdef LINEAR_SCALE
			for (j=0; j<3; j++) {
				newval = (int)((((double)(*(ptr+j))+0.5)*scaling) + rand1());
				if (newval > 255) {
					if (*(ptr+1) != 0) {
						check1 = check1 + 1.0;
					}
					newval = 255;
				}
				*(ptr+j) = (BYTE)newval;
			}
#else
			newval = *(ptr+0) + scaleaddB;
			if (newval > 255) newval = 255;
			else if (newval < 0) newval = 0;
			*(ptr+0) = (BYTE)newval;

			newval = *(ptr+1) + scaleaddG;
			if (newval > 255) newval = 255;
			else if (newval < 0) newval = 0;
			*(ptr+1) = (BYTE)newval;

			newval = *(ptr+2) + scaleaddR;
			if (newval > 255) newval = 255;
			else if (newval < 0) newval = 0;
			*(ptr+2) = (BYTE)newval;
#endif
			if (oldpb > MinPixBrightness) {
				modframe->brightness_after += CalcPixBrightness(ptr);
				numused++;
			}
			ptr += 3;
		}
		modframe->brightness_after /= (double)numused;
		delta = ((modframe->brightness_after-DesiredBrightness) / DesiredBrightness);
		if (delta < 0.0) delta = -delta;
		if (delta > 0.02) {
			check1 = modframe->brightness_after;
			check2 = modframe->percent_dark;
		}
	}
}

// This function can change depending on what kind of manipulation we're doing
// This function is responsible for eliminating samples we don't need any more
void Manipulator::ConsiderWriting() {
	// For this effort, we look backwards for a little over one film-frame and pick the brightest
	// recorded-frame; we manipulate the chosen recorded-frame as specified by
	// the user; and then we consider if we haven't inserted enough written-frames
	// for the written-frame rate, and if so, repeat the last one
	SIter	considerIter,nextIter,prevIter;
	int		i,frames_to_consider;

	if (WriteOutIter == AllSamples.end()) considerIter = AllSamples.begin();
	else {
		considerIter = WriteOutIter;
		for (i=0; i<FramesToSkipBeforeConsideration; i++) considerIter++;		// May fall off end
	}

	if (considerIter == AllSamples.end()) frames_to_consider = 0;
	else frames_to_consider = ReadInIter->framenum - considerIter->framenum;	// Note the just-read sample is always left alone

	if (frames_to_consider >= FramesNeededForConsideration) {
		// Okay, we've got enough frames to work with.  Find the one closest to the desired brightness.
		SIter	bestIter=considerIter;
		double	bestdelta,thisdelta;

		bestdelta=bestIter->brightness-DesiredBrightness;
		if (bestdelta<0.0) bestdelta=-bestdelta;
		for (i=0; i<FramesNeededForConsideration; i++) {
			considerIter++;
			thisdelta=considerIter->brightness-DesiredBrightness;
			if (thisdelta<0.0) thisdelta=-thisdelta;
			if (thisdelta<bestdelta) {
				bestIter = considerIter;
				bestdelta = thisdelta;
			}
		}

		// We have our magic frame --- the best example of a single film-frame.  Mess with the bitmap.
		ModifyBitmap(bestIter);
		// We're ready to write
		double frame_steps;
		if (WriteOutIter==AllSamples.end()) frame_steps=(double)(bestIter->framenum);
		else frame_steps = (double)(bestIter->framenum-WriteOutIter->framenum);

		if (frame_steps == 0) {
			//Huh?
			frame_steps = 0;
			if (bestIter == WriteOutIter) {
				//Huh??
				frame_steps = 0;
			}
			frame_steps = bestIter->framenum;
			frame_steps = WriteOutIter->framenum;
			considerIter = WriteOutIter;
			if (WriteOutIter == AllSamples.end()) {
				frame_steps = 0;
			}
			else if (WriteOutIter == AllSamples.begin()) {
				frame_steps = WriteOutIter->framenum;
			}
			for (i=0; i<FramesToSkipBeforeConsideration; i++) {
				considerIter++;
				frame_steps = considerIter->framenum;
			}
			for (i=0; i<FramesNeededForConsideration; i++) {
				considerIter++;
				frame_steps = considerIter->framenum;
			}
		}

		WriteFilmedTime += frame_steps * FilmTimePerRecordedFrame;
		WriteVideoTime += VideoTimePerWrittenFrame;
		WriteOutIter = bestIter;
		WriteFrame(WriteOutIter, WriteVideoTime);

		// Check our time information to see if we're running a frame behind
		// We assume this is the case, i.e. that video rates are faster than film rates
		if ((WriteFilmedTime - WriteVideoTime) > VideoTimePerWrittenFrame) {
			// Write the last frame again
			WriteVideoTime += VideoTimePerWrittenFrame;
			WriteFrame(WriteOutIter, WriteVideoTime);
		}

		// Eliminate the frames before the one just written
		for (considerIter=AllSamples.begin(); considerIter!=WriteOutIter; ) {	// Incrementing is done via erase
			SAFE_RELEASE(considerIter->sample);
			considerIter = AllSamples.erase(considerIter);
		}
	}
}


BYTE *Manipulator::CalcPtr(SIter frame, int x, int y) {
	BYTE	*ptr;
	ptr = frame->map + (y*frame->info->bmiHeader.biWidth*3) + (x*3);
	return(ptr);
}

double Manipulator::CalcVec(BYTE *ptr) {
	double	vec;
	double	bval,rval,gval;

	bval=(double)*ptr++;
	gval=(double)*ptr++;
	rval=(double)*ptr++;
	vec=RFAC*rval + GFAC*gval + BFAC*bval;
	return(vec);
}

double Manipulator::CalcVecDiff(BYTE *ptr1, BYTE *ptr2) {
	double	diff,vec1,vec2;
	double	bval1,rval1,gval1;
	double	bval2,rval2,gval2;
	double	bvald,rvald,gvald;

	bval1=(double)*ptr1++ * BFAC;
	gval1=(double)*ptr1++ * GFAC;
	rval1=(double)*ptr1++ * RFAC;
	bval2=(double)*ptr2++ * BFAC;
	gval2=(double)*ptr2++ * GFAC;
	rval2=(double)*ptr2++ * RFAC;
	vec1=sqrt(bval1*bval1 + gval1*gval1 + rval1*rval1);
	vec2=sqrt(bval2*bval2 + gval2*gval2 + rval2*rval2);
	bvald=bval2-bval1;
	gvald=gval2-gval1;
	rvald=rval2-rval1;
	diff=sqrt(bvald*bvald + gvald*gvald + rval2*rval2);
	if (vec1 > vec2) diff = -diff;
	return(diff);
}


void Manipulator::PreprocessBitmap(SIter modframe) {
#if 0
	// Test: perform a gradient filter
	int		pixx,pixy,checkx,checky,limx,disti;
	BYTE	*pixptr,*s1ptr,*s2ptr;
	double	cx,cy,dist;
	double	s1v,s2v,vgrad,vgradmax;
	BYTE	*newimg;

	newimg = new BYTE[modframe->info->bmiHeader.biHeight*modframe->info->bmiHeader.biWidth*3];
	vgradmax = 0.0;
	for (pixy=GradientReach; pixy<modframe->info->bmiHeader.biHeight-GradientReach; pixy++) {
		for (pixx=GradientReach; pixx<modframe->info->bmiHeader.biWidth-GradientReach; pixx++) {
			if ((pixx==150) && (pixy==100)) {
				dist = 0.0;		// Something to put a breakpoint on
			}
#if 0
			vgradmax = 0.0;
			for (checky=-GradientReach; checky<=GradientReach; checky++) {
				if (checky >= 0) limx = -1; else limx = 0;
				for (checkx=-GradientReach; checkx<=limx; checkx++) {
					cx=(double)checkx;
					cy=(double)checky;
					dist=sqrt((cx*cx)+(cy*cy));
					disti=(int)(dist+0.5);
					if (disti==GradientReach) {
						s1ptr=CalcPtr(modframe, pixx+checkx, pixy+checky);
						s2ptr=CalcPtr(modframe, pixx-checkx, pixy-checky);
						s1v=CalcVec(s1ptr);
						s2v=CalcVec(s2ptr);
						vgrad=s2v-s1v;
						if (vgrad < 0.0) vgrad = -vgrad;
						if (vgrad > vgradmax) vgradmax = vgrad;
					}
				}
			}
			pixptr=CalcPtr(modframe, pixx, pixy);
//			vgradmax/=CalcVec(pixptr);
//			vgradmax*=100.0;
			if (vgradmax < 256.0) {
				*pixptr++ = (BYTE)vgradmax;
				*pixptr++ = 0;
				*pixptr++ = 0;
			}
			else {
				*pixptr++ = 255;
				vgradmax -= 255.0;
				if (vgradmax < 255.0) {
					*pixptr++ = (BYTE)vgradmax;
					*pixptr++ = 0;
				}
				else {
					*pixptr++ = 255;
					vgradmax -= 255.0;
					if (vgradmax < 255.0) *pixptr++ = (BYTE)vgradmax;
					else *pixptr++ = 255.0;
				}
			}
#endif
#if 1
			s1ptr=CalcPtr(modframe, pixx-GradientReach, pixy);
			s2ptr=CalcPtr(modframe, pixx+GradientReach, pixy);
			vgrad=CalcVecDiff(s1ptr, s2ptr);
			pixptr = newimg + (pixy*modframe->info->bmiHeader.biWidth*3) + (pixx*3);
			if (vgrad >= 0.0) {
				if (vgrad <= 255.0)	*pixptr++ = (BYTE)vgrad;
				else *pixptr++ = 255;
				*pixptr++ = 0;
				*pixptr++ = 0;
				vgradmax += vgrad;
			}
			else {
				*pixptr++ = 0;
				*pixptr++ = 0;
				if (vgrad >= -255.0) *pixptr++ = (BYTE)(-vgrad);
				else *pixptr++ = 255;
				vgradmax -= vgrad;
			}
#endif
#if 0
			pixptr=CalcPtr(modframe, pixx, pixy);
			vgrad=CalcVec(pixptr);
			*pixptr++ = (BYTE)vgrad;
			*pixptr++ = (BYTE)vgrad;
			*pixptr++ = (BYTE)vgrad;
#endif
		}
	}
#if 1
	vgradmax /= (modframe->info->bmiHeader.biHeight-2*GradientReach)*(modframe->info->bmiHeader.biWidth-2*GradientReach);
	for (pixy=GradientReach; pixy<modframe->info->bmiHeader.biHeight-GradientReach; pixy++) {
		for (pixx=GradientReach; pixx<modframe->info->bmiHeader.biWidth-GradientReach; pixx++) {
			pixptr = newimg + (pixy*modframe->info->bmiHeader.biWidth*3) + (pixx*3);
			if (*pixptr > 1.2*vgradmax) *pixptr++ = 255; else *pixptr++ = 0;
			if (*pixptr > 1.2*vgradmax) *pixptr++ = 255; else *pixptr++ = 0;
			if (*pixptr > 1.2*vgradmax) *pixptr++ = 255; else *pixptr++ = 0;
		}
	}
#endif
	memcpy(modframe->map, newimg, modframe->info->bmiHeader.biHeight*modframe->info->bmiHeader.biWidth*3);
	delete [] newimg;
#endif
}


// This is the workhorse function: take in a new sample, keep it, maybe write things
void Manipulator::ModifyFrame(INSSBuffer *readSample, BITMAPINFO *readInfo, 
							  BITMAPINFO **dispInfo, BYTE **dispMap) {
	// Add this frame to the list
	ManipSample	readstruct;

	readstruct.sample = readSample;
	readstruct.info = readInfo;
	readSample->GetBuffer(&readstruct.map);
	readstruct.brightness = CalcBrightness(readstruct.info, readstruct.map, &readstruct.percent_dark);
	readstruct.brightness_after = 0.0;
	readstruct.framenum = ++last_framenum;

	AllSamples.push_back(readstruct);
	ReadInIter = AllSamples.end();
	ReadInIter--;

	// Do whatever we do with every frame
//	PreprocessBitmap(ReadInIter);

	// Now figure out if we want to do anything
	ConsiderWriting();

	// Send back whichever frame the user wants to watch
	if ((*DispIter) == AllSamples.end()) {
		// Haven't written anything yet
		*dispInfo = NULL;
		*dispMap = NULL;
	}
	else {
		*dispInfo = (*DispIter)->info;
		*dispMap = (*DispIter)->map;
	}
}


void Manipulator::NewParams(double inBright, double inMinPix, double inDarkFac) {
	DesiredBrightness = inBright;
	MinPixBrightness = inMinPix;
	DarknessFactor = inDarkFac;
}


