
#include "StdAfx.h"
#include "MovieWriter.h"
#include <list>
#include <vector>


using namespace std;


class Manipulator {
public:
	Manipulator();
	Manipulator(MovieWriter *writer_in);
	~Manipulator();

	void ModifyFrame(INSSBuffer *readSample, BITMAPINFO *readInfo, 
							  BITMAPINFO **dispInfo, BYTE **dispMap);

	void SetDisp(int whichdisp) {
		if (whichdisp == 1) {
			dispWrite = TRUE;
			DispIter = &WriteOutIter;
		}
		else {
			dispWrite = FALSE;
			DispIter = &ReadInIter;
		}
	}

	void SetManip(int howmanip) {
		if (howmanip == 1) ForceBrightness = TRUE;
		else ForceBrightness = FALSE;
	}

	void SetMarkMinPix(int mark) {
		if (mark == 1) MarkMinPix = TRUE;
		else MarkMinPix = FALSE;
	}

	double GetBrightness(double *outbright=NULL, double *percdark=NULL) { 
		if (outbright != NULL) *outbright = (*DispIter)->brightness_after;
		if (percdark != NULL) *percdark = (*DispIter)->percent_dark;
		return((*DispIter)->brightness);
	}

	void NewParams(double inBright, double inMinPix, double inDarkFac);

private:
	typedef struct {
		INSSBuffer	*sample;
		BITMAPINFO	*info;
		BYTE		*map;
		double		brightness;
		double		brightness_after;
		double		percent_dark;
		int			framenum;	// Can't do differences between iterators for some reason
	} ManipSample;

	typedef list<ManipSample>::iterator SIter;

	double CalcPixBrightness(BYTE *bmap);
	double CalcBrightness(BITMAPINFO *bmapinfo, BYTE *bmap, double *darkp=NULL);
	void AdjustParameters();
	void PreprocessBitmap(SIter modframe);
	void ModifyBitmap(SIter modframe);
	void ConsiderWriting();
	void WriteFrame(SIter towrite, double timewrite);
	void Setup();

	BYTE *CalcPtr(SIter frame, int x, int y);
	double CalcVec(BYTE *ptr);
	double CalcVecDiff(BYTE *ptr1, BYTE *ptr2);

	list<ManipSample> AllSamples;

	SIter	ReadInIter;
	SIter	WriteOutIter;
	SIter	*DispIter;

	BOOL dispWrite;

	MovieWriter	*mWriter;

	int		last_framenum;

	// Modification parameters
	double	FilmRecordFPS;			// Rate at which the film was originally recorded
	double	FilmPlaybackFPS;		// Rate at which the film was played back for video recording
	double	VideoRecordFPS;			// Rate at which the video recording of the film was made
	double	VideoPlaybackFPS;		// Rate at which the video is being written

	int		GradientReach;
	int		BrightnessRank;
	double	DesiredBrightness;
	BOOL	ForceBrightness;
	BOOL	MarkMinPix;
	double	MinPixBrightness;
	double	DarknessFactor;

	int		FramesToSkipBeforeConsideration;	// WriteOutIter is left at the last brightness peak; make sure we move past the other video frames for that film frame
	int		FramesNeededForConsideration;		// How many recorded frames to look at, more-or-less centered on the center of the film frame

	double	FilmTimePerRecordedFrame;			// How much film time a single recorded frame corresponds to
	double	VideoTimePerWrittenFrame;			// Essentially the inverse of the video frame time
	double	WriteFilmedTime;					// What the WriteOut pointer corresponds to in terms of film time
	double	WriteVideoTime;						// What the WriteOut pointer corresponds to in terms of video playback time
};


